###################################################################################################
#####                                                                                         #####
#####                                    ShiftML 2.0 model                                    #####
#####                     Author: Manuel Cordova (manuel.cordova@epfl.ch)                     #####
#####                                Last modified: 2022-03-16                                #####
#####                                    Please cite [doi]                                    #####
#####                                                                                         #####
###################################################################################################

import numpy as np
import os
import pickle as pk
import copy

import rascal
import rascal.representations as rep

# Setting the model directory
with open(os.path.dirname(os.path.abspath(__file__)) + "/model_dir.txt", "r") as F:
    _model_dir = F.read()



def load_model(elem, model_dir=None):
    """
    Load the model and representation of local environments
    
    Input:      - elem              Element to predict ("H", "C", "N", "O", "S", "F",
                                                        "P", "Cl", "Na", "Ca", "Mg" or "K")
    
    Outputs:    - X_train           Sparse representation of the training environments
                - ftr_inds_train    Indices associated with the sparse representation
                                        of the training environments
                - n_train           Total number of training environments
                - W                 Weight matrix for predictions
                - soap              SOAP descriptor of local atomic environment
                - zeta              Power of the kernel matrix
                - y_offset          Offset for normalization of computed shieldings
                - y_std             Standard deviation for normalization of computed shieldings
                - alpha             Rescaling factor for uncertainty prediction
    """
    
    if model_dir is not None:
        with open(f"{model_dir}ShiftML_v2_model_{elem}.pk", "rb") as F:
            [X_train, ftr_inds_train, n_train, W,
             rep_pars, zeta, y_offset, y_std, alpha] = pk.load(F)
    else:
        with open(f"{_model_dir}ShiftML_v2_model_{elem}.pk", "rb") as F:
            [X_train, ftr_inds_train, n_train, W,
             rep_pars, zeta, y_offset, y_std, alpha] = pk.load(F)
        
    soap = rep.SphericalInvariants(**rep_pars)
    
    return X_train, ftr_inds_train, n_train, W, soap, zeta, y_offset, y_std, alpha



def prepare_structures(structs, elem):
    """
    Prepare structures before constructing the representation of their atomic environments
    
    Inputs: - structs   Structure (or list of structures) to prepare
            - elem      Element to build the representation for

    Output: - structs   Prepared structure (or list of structures)
    
    """
    
    sml_structs = copy.deepcopy(structs)
    
    if isinstance(sml_structs, list):

        for struct in sml_structs:
            mask = [True if s == elem else False for s in struct.get_chemical_symbols() ]
            struct.set_array("center_atoms_mask",np.array(mask))
            struct.wrap(eps=1e-12)
            
    else:
        mask = [True if s == elem else False for s in sml_structs.get_chemical_symbols() ]
        sml_structs.set_array("center_atoms_mask",np.array(mask))
        sml_structs.wrap(eps=1e-12)
    
    return sml_structs



def build_rep(soap, structs):
    """
    Build the descriptor representation of atomic environments in a set of structures
    
    Inputs:     - soap          SOAP descriptor of local atomic environment
                - structs       Structure (or list of structures) to build the representation for
    
    Outputs:    - clean_X       Sparse descriptor representation of the atomic environments
                - ftr_inds      Indices where given features are present
                - n_samples     Total number of atomic environments constructed
    """

    X = soap.transform(structs).get_features_by_species(soap)

    ftr_inds = {}
    clean_X = {}
    for k in X:
        n_samples = X[k].shape[0]
        ftr_inds[k] = np.where(np.any(X[k] > 0., axis=1))[0]
        if len(ftr_inds[k]) > 0:
            clean_X[k] = X[k][ftr_inds[k]]
        else:
            clean_X[k] = None
    del X

    return clean_X, ftr_inds, n_samples



def build_representation(structs, elem, soap):
    """
    Build the descriptor representation of atomic environments of a given atom in a set of structures
    
    Inputs:     - structs       Structure (or list of structures) to build the representation for
                - elem          Element to build the representation for
                - soap          SOAP descriptor of local atomic environment
    
    Outputs:    - clean_X       Sparse descriptor representation of the atomic environments
                - ftr_inds      Indices where given features are present
                - n_samples     Total number of atomic environments constructed
    """
    
    structs = prepare_structures(structs, elem)
    
    return build_rep(soap, structs)



def build_kernel(X1, inds1, n1, X2, inds2, n2, zeta=1.):
    """
    Build the kernel between two sets of atomic environments
    
    Inputs: - X1        Sparse descriptor representation of the first set of atomic environments
            - inds1     Indices where given features are present in the first set of atomic environments
            - n1        Number of atomic environments in the first set
            - X2        Sparse descriptor representation of the second set of atomic environments
            - inds2     Indices where given features are present in the second set of atomic environments
            - n2        Number of atomic environments in the second set
            - zeta      Kernel power

    Output: - K         Kernel between the two sets of atomic environments
    """
    
    K = np.zeros((n2, n1))
    
    for k in X1:
        if len(inds1[k]) > 0 and len(inds2[k]) > 0:
            K[inds2[k][:, None], inds1[k]] += X2[k].dot(X1[k].T)
    
    return K ** zeta



def predict(X_train, ftr_inds_train, n_train, X_test, ftr_inds_test, n_test,
            W, y_offset, y_std, zeta=1., alpha=1.):
    """
    Predict target values on a test set using an ensemble of trained models
    
    Inputs:     - Xs_train      List of training representations for each model in the ensemble
                - X_test        Test representation
                - weights       List of weights for each model in the ensemble
                - y_offset      Offset for target values
                - y_std         Standard deviation of target values in the training data
                - zeta          Power to raise the kernel to
    
    Outputs:    - y_pred        Predicted values for the test set
                - y_err         Predicted errors for the test set
    """
    
    print("    Building kernel...")
    K = build_kernel(X_train, ftr_inds_train, n_train, X_test, ftr_inds_test, n_test, zeta=zeta)
    print("    Making predictions...")
    y_preds = (K.dot(W) * y_std) + y_offset
    
    # Compute mean and standard deviations of predictions
    y_pred = np.mean(y_preds, axis=1)
    y_err = np.std(y_preds, axis=1) * alpha
    
    return y_pred, y_err



###################################################################################################
#####                                                                                         #####
#####       The Model object below can be used as an alternative to the functions above       #####
#####                                                                                         #####
###################################################################################################



class Model():
    """
    ShiftML model class
    """
    
    def __init__(self, elem, model_dir=None):
        """
        Initialize a ShiftML model for the desired element
        """
        if model_dir is not None:
            self.mod_dir = model_dir
        else:
            self.mod_dir = _model_dir
            
        # Set element
        self.elem = elem
        
        # Load the model
        
        with open(f"{self.mod_dir}ShiftML_v2_model_{self.elem}.pk", "rb") as F:
            [self.X_train, self.ftr_inds_train, self.n_train, self.W,
             self.rep_pars, self.zeta, self.y_offset, self.y_std, self.alpha] = pk.load(F)
        
        self.soap = rep.SphericalInvariants(**self.rep_pars)
    
        return
    
    def prepare_structures(self, structs=None):
        """
        Prepare structures before constructing the representation of their atomic environments

        Input:  - structs   Structure (or list of structures) to prepare
        """
        
        if structs is not None:
            self.structs = copy.deepcopy(structs)

        if isinstance(self.structs, list):

            for struct in self.structs:
                mask = [True if s == self.elem else False for s in struct.get_chemical_symbols() ]
                struct.set_array("center_atoms_mask",np.array(mask))
                struct.wrap(eps=1e-12)

        else:
            mask = [True if s == self.elem else False for s in self.structs.get_chemical_symbols() ]
            self.structs.set_array("center_atoms_mask",np.array(mask))
            self.structs.wrap(eps=1e-12)

        return



    def build_rep(self):
        """
        Build the descriptor representation of atomic environments in a set of structures
        """

        X = self.soap.transform(self.structs).get_features_by_species(self.soap)

        ftr_inds = {}
        clean_X = {}
        for k in X:
            n_samples = X[k].shape[0]
            ftr_inds[k] = np.where(np.any(X[k] > 0., axis=1))[0]
            if len(ftr_inds[k]) > 0:
                clean_X[k] = X[k][ftr_inds[k]]
            else:
                clean_X[k] = None
        del X
        
        self.X_test = clean_X
        self.ftr_inds_test = ftr_inds
        self.n_test = n_samples

        return



    def build_representation(self, structs):
        """
        Build the descriptor representation of atomic environments of a given atom in a set of structures

        Inputs:     - structs       Structure (or list of structures) to build the representation for
                    - elem          Element to build the representation for
                    - soap          SOAP descriptor of local atomic environment

        Outputs:    - clean_X       Sparse descriptor representation of the atomic environments
                    - ftr_inds      Indices where given features are present
                    - n_samples     Total number of atomic environments constructed
        """

        self.structs = structs
        
        self.prepare_structures()
        self.build_rep()

        return



    def build_kernel(self):
        """
        Build the kernel between two sets of atomic environments
        """

        K = np.zeros((self.n_test, self.n_train))

        for k in self.X_train:
            if len(self.ftr_inds_train[k]) > 0 and len(self.ftr_inds_test[k]) > 0:
                K[self.ftr_inds_test[k][:, None], self.ftr_inds_train[k]] += self.X_test[k].dot(self.X_train[k].T)

        return K ** self.zeta



    def predict(self):
        """
        Predict target values on a test set using an ensemble of trained models

        Inputs:     - Xs_train      List of training representations for each model in the ensemble
                    - X_test        Test representation
                    - weights       List of weights for each model in the ensemble
                    - y_offset      Offset for target values
                    - y_std         Standard deviation of target values in the training data
                    - zeta          Power to raise the kernel to

        Outputs:    - y_pred        Predicted values for the test set
                    - y_err         Predicted errors for the test set
        """

        print("    Building kernel...")
        K = self.build_kernel()
        print("    Making predictions...")
        y_preds = (K.dot(self.W) * self.y_std) + self.y_offset

        # Compute mean and standard deviations of predictions
        y_pred = np.mean(y_preds, axis=1)
        y_err = np.std(y_preds, axis=1) * self.alpha

        return y_pred, y_err
